import sys

from black import main

sys.path.insert(0, ".")

from pigit.interaction.status_interaction import InteractiveStatus
from pigit.interaction.commit_interaction import InteractiveCommit


def status():
    # Initialize to debug mode.
    i = InteractiveStatus(debug=False)
    i.run()


def commit():
    i = InteractiveCommit()
    i.run()


commit()



import time
import textwrap
import pytest

from pigit.interaction.base import DataHandle
from pigit.interaction.status_interaction import InteractiveStatus


@pytest.fixture(scope="module")
def setup():
    return InteractiveStatus(), DataHandle(use_color=True)


def test_files_time(setup):
    ia, dh = setup

    status = dh.load_status(100)
    assert type(status) == list

    print("\nWill get 1000 times data.")
    sum_time = max_time = 0
    min_time = float("inf")
    for _ in range(1000):
        start = time.time()
        dh.load_status(100)
        used_time = time.time() - start
        min_time = min(used_time, min_time)
        max_time = max(used_time, max_time)
        sum_time += used_time

    print(
        textwrap.dedent(
            f"""
            test [get_status] 1000 times:
                Total spend time: {sum_time}
                Min spend time: {min_time}
                Max spend time: {max_time}
            """
        )
    )



import os
import logging
import textwrap
from distutils.util import strtobool
from typing import Optional

from .common import confirm, is_color
from .common.singleton import Singleton

Log = logging.getLogger(__name__)


class ConfigError(Exception):
    """Config error. Using by `Config`."""

    pass


class Config(object, metaclass=Singleton):
    """PIGIT configuration class."""

    CONFIG_TEMPLATE: str = textwrap.dedent(
        """\
        #? Config file for pigit v. {version}

        #  ____ ___ ____ ___ _____                            __ _
        # |  _ \\_ _/ ___|_ _|_   _|           ___ ___  _ __  / _(_) __ _
        # | |_) | | |  _ | |  | |_____ _____ / __/ _ \\| '_ \\| |_| |/ _` |
        # |  __/| | |_| || |  | |_____|_____| (_| (_) | | | |  _| | (_| |
        # |_|  |___\\____|___| |_|            \\___\\___/|_| |_|_| |_|\\__, |
        #                                     {version:>20} |___/
        # Git-tools -- pigit configuration.

        # Show original git command.
        gitprocessor_show_original={gitprocessor_show_original}

        # Is it recommended to correct when entering wrong commands.
        gitprocessor_use_recommend={gitprocessor_use_recommend}

        # Whether color is enabled in interactive mode.
        gitprocessor_interactive_color={gitprocessor_interactive_color}

        # Display time of help information in interactive mode, 0 is permanent.
        gitprocessor_interactive_help_showtime={gitprocessor_interactive_help_showtime}

        # Whether to use the ignore configuration of the `.gitignore` file.
        codecounter_use_gitignore={codecounter_use_gitignore}

        # Whether show files that cannot be counted.
        codecounter_show_invalid={codecounter_show_invalid}

        # Whether show files icons. Font support required, like: 'Nerd Font'
        codecounter_show_icon={codecounter_show_icon}

        # Output format of statistical results.
        # Supported: [table, simple]
        # When the command line width is not enough, the `simple ` format is forced.
        codecounter_result_format={codecounter_result_format}

        # Timeout for getting `.gitignore` template from net.
        gitignore_generator_timeout={gitignore_generator_timeout}

        # Git local config print format.
        # Supported: [table, normal]
        git_config_format={git_config_format}

        # Control which parts need to be displayed when viewing git repository information.
        repository_show_path={repository_show_path}
        repository_show_remote={repository_show_remote}
        repository_show_branchs={repository_show_branchs}
        repository_show_lastest_log={repository_show_lastest_log}
        repository_show_summary={repository_show_summary}

        # Whether with color when use `-h` get help message.
        help_use_color={help_use_color}

        # The max line width when use `-h` get help message.
        help_max_line_width={help_max_line_width}

        # Whether run PIGIT in debug mode.
        debug_mode={debug_mode}

        # Whether output log in terminal.
        stream_output_log={stream_output_log}

        """
    )

    # yapf: disable
    _keys :list[str]= [
        'gitprocessor_show_original', 'gitprocessor_use_recommend',
        'gitprocessor_interactive_color', 'gitprocessor_interactive_help_showtime',
        'codecounter_use_gitignore', 'codecounter_show_invalid',
        'codecounter_result_format', 'codecounter_show_icon',
        'gitignore_generator_timeout',
        'git_config_format',
        'repository_show_path', 'repository_show_remote', 'repository_show_branchs',
        'repository_show_lastest_log', 'repository_show_summary',
        'help_use_color', 'help_max_line_width',
        'debug_mode', 'stream_output_log'
    ]
    # yapf: enable

    # config default values.
    gitprocessor_show_original: bool = True
    gitprocessor_use_recommend: bool = True
    gitprocessor_interactive_color: bool = True
    gitprocessor_interactive_help_showtime: float = 1.5

    codecounter_use_gitignore: bool = True
    codecounter_show_invalid: bool = False
    codecounter_show_icon: bool = False
    codecounter_result_format: str = "table"  # table, simple
    _supported_result_format: list = ["table", "simple"]

    gitignore_generator_timeout: int = 60

    git_config_format: str = "table"
    _supported_git_config_format: list = ["normal", "table"]

    repository_show_path: bool = True
    repository_show_remote: bool = True
    repository_show_branchs: bool = True
    repository_show_lastest_log: bool = True
    repository_show_summary: bool = False

    help_use_color: bool = True
    help_max_line_width: int = 90

    debug_mode: bool = False
    stream_output_log: bool = False

    # Store warning messages.
    warnings: list = []

    def __init__(self, path, current_version: str = "unknown") -> None:
        super(Config, self).__init__()

        self.config_path = path
        self.current_version = current_version

        conf = self.load_config()

        for key in self._keys:
            if key in conf.keys() and conf[key] != "==error==":
                setattr(self, key, conf[key])

    def load_config(self) -> dict:
        new_config = {}
        config_file = self.config_path
        try:
            with open(config_file) as cf:
                for line in cf:
                    line = line.strip()
                    if line.startswith("#? Config"):
                        new_config["version"] = line[line.find("v. ") + 3 :]
                        continue
                    if line.startswith("#"):
                        # comment line.
                        continue
                    if "=" not in line:
                        # invalid line.
                        continue
                    key, line = line.split("=", maxsplit=1)

                    # processing.
                    key = key.strip()
                    line = line.strip().strip('"')

                    # checking.
                    if key not in self._keys:
                        self.warnings.append("'{0}' is not be supported!".format(key))
                        continue
                    elif type(getattr(self, key)) == int:
                        try:
                            new_config[key] = int(line)
                        except ValueError:
                            self.warnings.append(
                                'Config key "{0}" should be an integer!'.format(key)
                            )
                    elif type(getattr(self, key)) == bool:
                        try:
                            # True values are y, yes, t, true, on and 1;
                            # false values are n, no, f, false, off and 0.
                            # Raises ValueError if val is anything else.
                            new_config[key] = bool(strtobool(line))
                        except ValueError:
                            self.warnings.append(
                                'Config key "{0}" can only be True or False!'.format(
                                    key
                                )
                            )
                    elif type(getattr(self, key)) == str:
                        if "color" in key and not is_color(line):
                            self.warnings.append(
                                'Config key "{0}" should be RGB string, like: #FF0000'.format(
                                    key
                                )
                            )
                            continue
                        new_config[key] = str(line)
        except Exception as e:
            Log.error(str(e) + str(e.__traceback__))

        if (  # check codecounter output format whether supported.
            "codecounter_result_format" in new_config
            and new_config["codecounter_result_format"]
            not in self._supported_result_format
        ):
            new_config["codecounter_result_format"] = "==error=="
            self.warnings.append(
                'Config key "{0}" support must in {1}'.format(
                    "codecounter_result_format", self._supported_result_format
                )
            )

        if (
            "git_config_format" in new_config
            and new_config["git_config_format"] not in self._supported_git_config_format
        ):
            new_config["git_config_format"] = "==error=="
            self.warnings.append(
                'Config key "{0}" support must in {1}'.format(
                    "git_config_format", self._supported_git_config_format
                )
            )

        version = self.current_version
        if "version" in new_config and not (
            # If unknown current version or
            # If the version is right or
            # If current version is a [beta, alpha, dev] verstion then will not tip.
            # Else if version is not right will tip.
            version == "unknown"
            or new_config["version"] == version
            or "beta" in version
            or "alpha" in version
            or "dev" in version
        ):
            print(new_config["version"])
            self.warnings.append(
                "The current configuration file is not up-to-date."
                "You'd better recreate it."
            )

        return new_config

    def create_config_template(self) -> None:
        parent_dir = os.path.dirname(self.config_path)
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir, exist_ok=True)

        if os.path.exists(self.config_path) and not confirm(
            "Configuration exists, overwrite? [y/n]"
        ):
            return None

        # Try to load already has config.
        conf = self.load_config()
        for key in self._keys:
            if not conf.get(key, None) or conf[key] == "==error==":
                conf[key] = getattr(self, key)
        conf["version"] = self.current_version

        # Write config. Will save before custom setting.
        try:
            with open(
                self.config_path, "w" if os.path.isfile(self.config_path) else "x"
            ) as f:
                f.write(self.CONFIG_TEMPLATE.format(**conf))
        except Exception as e:
            Log.error(str(e) + str(e.__traceback__))
            print("Failed, create config.")
        else:
            print("Successful.")



# -*- coding:utf-8 -*-
import os

# For windows print color.
if os.name == "nt":
    os.system("")

from .emotion import Emotion
from .style import Color, BoxSymbol
from .escape import Fx, Cursor
from .utils import run_cmd, exec_cmd, confirm, color_print, is_color, similar_command
from .str_utils import get_width, shorten, adjudgment_type, get_file_icon


# yapf: disable
__all__ = [
    "Emotion", "Color", "BoxSymbol", "Fx", "Cursor",
    "run_cmd", "exec_cmd", "confirm", "color_print", "is_color", "similar_command",
    "get_width", "shorten", "adjudgment_type", "get_file_icon",
]
# yapf: enable


class TermColor:
    """Terminal print color class."""

    Red = Color.fg("#FF6347")  # Tomato
    Green = Color.fg("#98FB98")  # PaleGreen
    DeepGreen = Color.fg("#A4BE8C")  # PaleGreen
    Yellow = Color.fg("#EBCB8C")
    Gold = Color.fg("#FFD700")  # Gold
    SkyBlue = Color.fg("#87CEFA")
    MediumVioletRed = Color.fg("#C71585")
    Symbol = {"+": Color.fg("#98FB98"), "-": Color.fg("#FF6347")}



import argparse
import pytest
from pprint import pprint
from .utils import analyze_it

from pigit import Git_Cmds, __project__, Parser
from pigit.shell_completion import ShellCompletion, process_argparse


class TestCompletion:
    def test_error_complete_vars(self):
        with pytest.raises(TypeError):
            ShellCompletion("test", "xxx", ".")

    def test_error_shell(self):
        with pytest.raises(ValueError):
            ShellCompletion("test", {}, ".", shell="cmd")

    def test_error_argparse_obj(self):
        with pytest.raises(TypeError):
            process_argparse("xxx")

        with pytest.raises(TypeError):
            process_argparse(object)

    def test_parse_parser(self):
        p = Parser()
        p.parse([""])
        res = process_argparse(p._parser)
        pprint(res)

    @analyze_it
    def test_generater(self):

        for item in ShellCompletion.Supported_Shell:
            print(item)
            complete_vars = {
                key: value.get("help", "") for key, value in Git_Cmds.items()
            }
            handle = ShellCompletion(
                __project__, complete_vars, shell=item, script_dir="."
            )
            pprint(handle.generate_resource())



# -*- coding:utf-8 -*-

import os
import re
import logging

from .utils import exec_cmd

Log = logging.getLogger(__name__)


def git_version() -> str:
    """Get Git version."""

    _, git_version_ = exec_cmd("git --version")
    Log.debug("Detect git version:" + str(git_version_))
    return git_version_ or ""


def current_repository() -> tuple[str, str]:
    """
    Get the current git repository path. If not, the path is empty.
    Get the local git config path. If not, the path is empty.

    Return:
        (tuple[str, str]): repository path, git config path.
    """

    err, path = exec_cmd("git rev-parse --git-dir")

    repo_path: str = ""
    git_conf_path: str = ""

    if err:
        return repo_path, git_conf_path

    # remove useless space.
    path = path.strip()

    if ".git/submodule/" in path:
        # this repo is submodule.
        git_conf_path = path
        repo_path = path.replace(".git/submodule/", "")
    if path == ".git":
        repo_path = os.getcwd()
        git_conf_path = os.path.join(repo_path, ".git")
    else:
        git_conf_path = path
        repo_path = path[:-5]

    Log.debug("Final repo: {0}, {1}".format(repo_path, git_conf_path))
    return repo_path, git_conf_path


def parse_git_config(conf: str) -> dict:
    conf = re.split(r"\r\n|\r|\n", conf)
    config_dict: dict[str, dict[str, str]] = {}
    config_type: str = ""

    for line in conf:
        line = line.strip()

        if not line:
            continue

        elif line.startswith("["):
            config_type = line[1:-1].strip()
            config_dict[config_type] = {}

        elif "=" in line:
            key, value = line.split("=", 1)
            config_dict[config_type][key.strip()] = value.strip()

        else:
            continue

    # debug info.
    Log.debug(config_dict)

    return config_dict


if __name__ == "__main__":
    from pprint import pprint

    conf = """
    [core]
        repositoryformatversion = 0
        filemode = true
        bare = false
        logallrefupdates = true
        ignorecase = true
        precomposeunicode = true
    [remote "origin"]
        url = https://github.com/zlj-zz/pigit.git
        fetch = +refs/heads/*:refs/remotes/origin/*
    [branch "main"]
        remote = origin
        merge = refs/heads/main
    [branch "pygittolls"]
        remote = origin
        merge = refs/heads/pygittolls
    [branch "pygittools"]
        remote = origin
        merge = refs/heads/pygittools
    [credential]
        helper = store
    [branch "split_file"]
        remote = origin
        merge = refs/heads/split_file
    [branch "singla_file"]
        remote = origin
        merge = refs/heads/singla_file
    [branch "compat-for-2"]
        remote = origin
        merge = refs/heads/compat-for-2
    """
    pprint(parse_git_config(conf))



import time
import textwrap
import pytest

from pigit.common.str_table import Table, dTable


@pytest.mark.parametrize(
    "header, data, title",
    [
        (
            ["name", "age", "gender"],
            [
                ["bob", "20", "f"],
                ["tom", "19", "f"],
            ],
            "Student",
        ),
        (
            ["name", "age", "gender"],
            [
                ["bob", "20", "\033[31;1mf\033[0m"],
                ["tom", "19", "f"],
            ],
            "Has missing",
        ),
        (
            ["name", "age", "gender"],
            [
                ["bob", "20", "f"],
                ["tom", "19"],
            ],
            "",
        ),
        (
            ["名字", "年龄", "性别"],
            [
                ["张三", "20", "f"],
                ["李四", "19", "f"],
            ],
            "",
        ),
    ],
)
def test_table1(header, data, title):
    print()
    table = Table(header, data, title)
    table.print()
    print(table.each_column_width)
    # print(table)


@pytest.mark.parametrize(
    "d_data",
    [
        {
            "Fruit color": {
                "apple": "red",
                "grape": "purple",
            },
            "Animal color": {
                "cattle": "yellow",
                "sheep": "white",
            },
        },
        {
            "File suffixes for some languages": {
                "C": "c",
                "C#": "csharp",
                "Python": "py",
                "JavaScript": "js",
            },
        },
        {
            "Fruit color": {
                "apple": "\033[31mred\033[0m",
                "grape": "purple",
            },
        },
        {
            "水果颜色": {
                "苹果": "红色",
                "葡萄": "紫色",
            },
        },
    ],
)
def test_table2(d_data):
    print()
    table = dTable(d_data)
    table.print()



import os
import time
import threading
import pytest

import concurrent.futures
from pprint import pprint
from .utils import analyze_it

from pigit import COUNTER_PATH
from pigit.codecounter import CodeCounter, CodeCounterError


@analyze_it
def test_codecounter(path=os.getcwd()):
    with pytest.raises(CodeCounterError):
        CodeCounter(result_format="xxx")

    start_t = time.time()
    CodeCounter(
        result_saved_path=COUNTER_PATH, count_path=path
    ).count_and_format_print()
    print(time.time() - start_t)


@pytest.mark.parametrize(
    "path",
    [
        os.getcwd(),
        # *os.environ.values(),
    ],
)
def test_pure_walk(path):
    print(f"\n===Test pure walk ({path})===")
    count = 0
    start_t = time.time()
    for root, dirs, files in os.walk(path):
        for _ in files:
            count += 1
            # print("\rFound files: {0}".format(count), end="")
    print("")
    print(f"Result spend time: {time.time() - start_t}")


"""
[just match root]
13.6025s

[record lines count]
use thread: 65.19s
no thread:  77.51s

[print content]
use thread: 570.698s
no thread:  569.579s

877.2
954.35
"""


def _read_handle(root, files, counter):
    # print(counter.Rules)
    result_count = 0

    # First, filter folder.
    is_effective_dir = counter.matching(root)
    if not is_effective_dir:
        return None

    for file in files:
        full_path = os.path.join(root, file)
        is_effective = counter.matching(full_path)
        if is_effective:
            # threading.Thread(target=_read_handle, args=(full_path,)).start()

            try:
                # Try read size of the valid file. Then do sum calc.
                size_ = os.path.getsize(full_path)
            except:
                pass
            try:
                with open(full_path) as f:
                    # do nothing.
                    result_count += len(f.read().split())
            except:
                pass
            pass
    return result_count


def test_mutliprocess_walk(path=os.getcwd()):
    print("\n===Test use mutliprocess walk===")
    progress = 0

    ignore_dir = []
    counter = CodeCounter()

    res_count = 0

    def _callback(r):
        nonlocal res_count
        res = r.result()
        if res:
            res_count += res

    start_t = time.time()

    with concurrent.futures.ProcessPoolExecutor() as pool:
        for root, dirs, files in os.walk(path):
            # Flag for large dir.
            progress += 1
            # print("\r{0}".format(progress), end="")

            # Process .gitignore to add new rule.
            if ".gitignore" in files:
                counter.process_gitignore(root)

            future_result = pool.submit(_read_handle, root, files, counter)
            future_result.add_done_callback(_callback)

        print("{0}".format(progress), end="")

        """
        # First, filter folder.
        is_effective_dir: bool = counter.matching(root)
        if not is_effective_dir:
            ignore_dir.append(root)
            continue

        for file in files:
            full_path: str = os.path.join(root, file)
            is_effective: bool = counter.matching(full_path)
            if is_effective:

                try:
                    # Try read size of the valid file. Then do sum calc.
                    size_ = os.path.getsize(full_path)
                except:
                    pass
                try:
                    with open(full_path) as f:
                        # do nothing.
                        len(f.read().split())
                except:
                    pass
                pass
        """
    print("\nover")
    # pprint(counter.Rules)
    # print(ignore_dir)
    print(time.time() - start_t)
    print("total:", res_count)


if __name__ == "__main__":
    # test_regular_rule()
    # test_regular_rule(os.environ["HOME"] + "/.config")
    # test_regular_rule(os.environ["HOME"] + "/prog")
    # test_regular_rule(os.environ["HOME"])
    # test_pure_walk(os.environ["HOME"] + "/prog")
    # test_pure_walk(os.environ["HOME"])

    test_codecounter(os.environ["HOME"] + "/prog")

    pass



# -*- coding:utf-8 -*-

from abc import ABC, abstractmethod
from shutil import get_terminal_size
from copy import deepcopy
from typing import Generator

from .style import BoxSymbol
from .escape import Fx
from .str_utils import get_width


class TableTooWideError(Exception):
    pass


class _baseTable(ABC):

    """Docstring for baseTable."""

    each_column_width: list

    def __init__(self, frame_format: str, nil: str, title: str = ""):
        if frame_format not in BoxSymbol.rune.keys():
            frame_format = "bold"
        self.rune = BoxSymbol.rune[frame_format]
        self.nil = nil
        self.title = title

        # create table when init.
        self.fix_data()

    @property
    def display_title(self) -> str:
        if self.title:
            line_max = sum(self.each_column_width) + len(self.each_column_width) + 1
            _t = "{}{:^%s}{}" % line_max
            return _t.format(Fx.b, self.title, Fx.ub)
        return "\r"

    @property
    def row_width(self) -> int:
        """Gets the width of table."""
        return sum(self.each_column_width) + len(self.each_column_width) + 1

    def real_len(self, text: str):
        """Gets the true width of the element on the command line."""
        return sum([get_width(ord(ch)) for ch in text])

    @abstractmethod
    def fix_data(self):
        """
        Calculate the width of each column in the table, and realize
        ``each_column_width`` in the real sense.
        There may be different calculation methods for different forms
        of tables, so this method has subclass implementation.
        """

    @abstractmethod
    def table_generator(self) -> Generator:
        """
        Returns a table generator that is used to effectively generate
        a table by row instead of returning all at once.
        There may be different calculation methods for different forms
        of tables, so this method has subclass implementation.

        Yields:
            Generator
        """

    def print(self) -> None:
        """Output the table."""

        term_width, _ = get_terminal_size()
        if term_width < self.row_width:
            raise TableTooWideError("Terminal is not wide enough.")

        print(self.display_title)

        g = self.table_generator()
        for i in g:
            print(i, end="")


class Table(_baseTable):
    """Create a table from list.

    data format:
        header = ["name", "age", "gender"]
        data = [
            ["bob", "20", "f"],
            ["tom", "19", "f"],
        ]
        tb = Table(header, data)
        tb.print()

    table format:
        ┏━━━━┳━━━┳━━━━━━┓
        ┃name┃age┃gender┃
        ┣━━━━╋━━━╋━━━━━━┫
        ┃bob ┃20 ┃f     ┃
        ┃tom ┃19 ┃f     ┃
        ┗━━━━┻━━━┻━━━━━━┛

    """

    def __init__(
        self,
        header: list,
        data: list[list],
        title: str = "",
        frame_format: str = "bold",
        nil: str = "",
    ):
        # Check data.
        if not isinstance(header, list):
            raise TypeError("title need is a list.")
        self.header = deepcopy(header)

        # Check data.
        if not isinstance(data, list):
            raise TypeError("data need is a list.")
        for item in data:
            if not isinstance(item, list):
                raise TypeError("each item of data need is a list.")
        self.data = deepcopy(data)

        self.header_len = len(self.header)
        self.each_column_width = [self.real_len(i) for i in self.header]

        super().__init__(frame_format, nil, title=title)

    def fix_data(self):
        header_len = self.header_len

        for item in self.data:
            # Complete missing element.
            item_len = len(item)
            if item_len < header_len:
                item.extend([self.nil] * (header_len - item_len))
            elif item_len > header_len:
                item = item[0:header_len]
            # Calc each max.
            self._adjust_each_max(item)

    def add_row(self, row: list) -> None:
        # XXX: whether need deepcopy
        row_len = len(row)
        if row_len < self.header_len:
            row.extend([self.nil] * (self.header_len - row_len))
        elif row_len > self.header_len:
            row = row[0 : self.header_len]

        self._adjust_each_max(row)
        self.data.append(row)

    def _adjust_each_max(self, cells: list) -> None:
        for i, x in enumerate(cells):
            x_len = self.real_len(Fx.pure(x))
            self.each_column_width[i] = max(self.each_column_width[i], x_len)

    def table_generator(self) -> Generator:
        each_column_width = self.each_column_width
        rune = self.rune
        indexes = range(self.header_len)

        # top and title line.
        yield f"{rune[2]}{rune[-3].join([rune[0] * i for i in each_column_width])}{rune[3]}\n"
        yield rune[1]
        for idx in indexes:
            width = each_column_width[idx]
            cell = self.header[idx]
            cell_len = self.real_len(Fx.pure(cell))
            yield cell + " " * (width - cell_len) + rune[1]
        yield "\n"
        yield f"{rune[6]}{rune[-1].join([rune[0] * i for i in each_column_width])}{rune[7]}\n"

        # all rows.
        for cells in self.data:
            yield rune[1]
            for idx in indexes:
                width = each_column_width[idx]
                cell = cells[idx]
                cell_len = self.real_len(Fx.pure(cell))
                yield cell + " " * (width - cell_len) + rune[1]
            yield "\n"

        # bottom
        yield rune[4] + rune[-2].join([rune[0] * i for i in each_column_width]) + rune[
            5
        ] + "\n"


class dTable(_baseTable):
    """Create table from a special format dict.

    d_data format: dict[str, dict[str, str]]
        d_data = {
            'Fruit color': {
                'apple': 'red',
                'grape': 'purple',
            },
            'Animal color': {
                'cattle': 'yellow',
                'sheep': 'white',
            },
        }

    table format:
        ┏━━━━━━━━━━━━━┓
        ┃ Fruit color ┃
        ┣━━━━━━┳━━━━━━┫
        ┃apple ┃red   ┃
        ┃grape ┃purple┃
        ┣━━━━━━┻━━━━━━┫
        ┃Animal color ┃
        ┣━━━━━━┳━━━━━━┫
        ┃cattle┃yellow┃
        ┃sheep ┃white ┃
        ┣━━━━━━┻━━━━━━┫
        ┃      ────END┃
        ┗━━━━━━━━━━━━━┛

    """

    def __init__(
        self, d_data: dict, title: str = "", frame_format: str = "bold", nil: str = ""
    ):
        # check data whether right.
        if not isinstance(d_data, dict):
            raise TypeError("d_data need is a dict.")
        for item in d_data.values():
            if not isinstance(item, dict):
                raise TypeError("the item of d_data need is a dict.")

        self.data = deepcopy(d_data)

        super().__init__(frame_format, nil, title=title)

    def fix_data(self):
        self.each_column_width = each_column_width = [0, 0]
        max_subtitle_len = 0
        r_len = self.real_len

        for subtitle, sub_dict in self.data.items():
            max_subtitle_len = max(max_subtitle_len, r_len(subtitle))
            for k, v in sub_dict.items():
                each_column_width[0] = max(each_column_width[0], r_len(Fx.pure(k)))
                each_column_width[1] = max(each_column_width[1], r_len(Fx.pure(v)))

        # For ensure that the table is output correctly when the len of sub title
        # bigger than the len of item.
        sum_each_max = sum(each_column_width)
        if max_subtitle_len > sum_each_max:
            each_column_width[1] += max_subtitle_len - sum_each_max

    def table_generator(self) -> Generator:
        rune = self.rune
        each_column_width = self.each_column_width
        row_width = self.row_width
        r_len = self.real_len

        _end_template = "%(flag)s{:>%(number)s}%(flag)s\n" % {
            "flag": rune[1],
            "number": row_width - 2,
        }

        sub_top = f"{rune[6]}{rune[-3].join([rune[0] * i for i in each_column_width])}{rune[7]}\n"
        sub_bottom = f"{rune[6]}{rune[-2].join([rune[0] * i for i in each_column_width])}{rune[7]}\n"

        # top
        yield f"{rune[2]}{rune[0] * (row_width - 2)}{rune[3]}\n"

        for subtitle, sub_dict in self.data.items():
            # subtitle part
            yield rune[1]
            subtitle_len = r_len(Fx.pure(subtitle))
            _div, _mod = divmod(row_width - 2 - subtitle_len, 2)
            yield f"{_div * ' '}{subtitle}{(_div + _mod) * ' '}{rune[1]}\n"

            # sub dict
            yield sub_top
            for k, v in sub_dict.items():
                k_len = r_len(Fx.pure(k))
                yield rune[1] + k + (each_column_width[0] - k_len) * " " + rune[1]
                v_len = r_len(Fx.pure(v))
                yield v + (each_column_width[1] - v_len) * " " + rune[1] + "\n"
            yield sub_bottom

        # bottom
        yield _end_template.format("────END")
        yield f"{rune[4]}{rune[0] * (row_width - 2)}{rune[5]}\n"



from .utils import analyze_it

from pigit import Parser, introduce
from pigit.gitinfo import output_repository_info, output_git_local_config


@analyze_it
def test_info():
    introduce()

    output_repository_info()

    output_git_local_config("xxx")
    output_git_local_config("normal")
    output_git_local_config("table")


def test_show_help():
    p = Parser()
    p.parse([""])
    p._parser.print_help()



# -*- coding:utf-8 -*-

# MIT License
#
# Copyright (c) 2021 Zachary
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


__project__ = "pigit"
__version__ = "1.3.2"
__url__ = "https://github.com/zlj-zz/pigit.git"
__uri__ = __url__

__author__ = "Zachary Zhang"
__email__ = "zlj19971222@outlook.com"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2021 Zachary"


import os
import sys
import argparse
import logging
from shutil import get_terminal_size
from typing import Optional, Union

from .log import LogHandle
from .common import Color, Fx, TermColor, color_print
from .gitinfo import (
    Git_Version,
    REPOSITORY_PATH,
    output_repository_info,
    output_git_local_config,
)
from .decorator import time_it
from .config import Config
from .codecounter import CodeCounter
from .shell_completion import ShellCompletion, process_argparse
from .gitignore import GitignoreGenetor
from .interaction import InteractiveStatus, InteractiveCommit
from .processor import CmdProcessor, Git_Cmds, CommandType


Log = logging.getLogger(__name__)

#####################################################################
# Part of compatibility.                                            #
# Handled the incompatibility between python2 and python3.          #
#####################################################################

# For windows.
USER_HOME: str = ""
PIGIT_HOME: str = ""
IS_WIN: bool = sys.platform.lower().startswith("win")
Log.debug("Runtime platform is windows: {0}".format(IS_WIN))

if IS_WIN:
    USER_HOME = os.environ["USERPROFILE"]
    PIGIT_HOME = os.path.join(USER_HOME, __project__)
else:
    USER_HOME = os.environ["HOME"]
    PIGIT_HOME = os.path.join(USER_HOME, ".config", __project__)

LOG_PATH: str = PIGIT_HOME + "/log/{0}.log".format(__project__)
CONFIG_PATH: str = PIGIT_HOME + "/pigit.conf"
COUNTER_PATH: str = PIGIT_HOME + "/Counter"


#####################################################################
# Configuration.                                                    #
#####################################################################
CONFIG = Config(path=CONFIG_PATH, current_version=__version__)
if CONFIG.warnings:
    for warning in CONFIG.warnings:
        print(warning)
    CONFIG.warnings = []


#####################################################################
# Implementation of additional functions.                           #
#####################################################################
def introduce() -> None:
    """Print the description information."""

    # Print version.
    print(
        """
 ____ ___ ____ ___ _____
|  _ \\_ _/ ___|_ _|_   _|
| |_) | | |  _ | |  | |
|  __/| | |_| || |  | |
|_|  |___\\____|___| |_| version: {}
""".format(
            __version__
        )
    )

    # Print git version.
    if Git_Version is None:
        color_print("Don't found Git, maybe need install.", TermColor.Red)
    else:
        print(Git_Version)

    # Print package path.
    color_print("Local path: ", Fx.b, end="")
    f_p = __file__.replace("./", "")  # f_p -> file path
    color_print("%s\n" % f_p, TermColor.SkyBlue, Fx.underline)

    # Print description.
    color_print("Description:", Fx.b)
    color_print(
        "  Terminal tool, help you use git more simple."
        " Support Linux, MacOS and Windows.\n"
        "  It use short command to replace the original command, like: \n"
        "  ``{green}pigit ws{rs}`` -> ``{green}git status --short{rs}``,"
        " ``{green}pigit b{rs}`` -> ``{green}git branch{rs}``.\n"
        "  Also you use ``{green}pigit -s{rs}`` to get the all short command,"
        " have fun and good lucky.\n"
        "  The open source path on github: {url}".format(
            url=TermColor.SkyBlue + Fx.underline + __url__,
            green=TermColor.Green,
            rs=Fx.reset,
        ),
        Fx.italic,
    )

    print(
        "\nYou can use {green}-h{rs} or {green}--help{rs} "
        "to get help and more usage.".format(green=TermColor.Green, rs=Fx.reset)
    )


def get_extra_cmds() -> dict:
    """Get custom cmds.

    Load the `extra_cmds.py` file under PIGIT HOME, check whether `extra_cmds`
    exists, and return it. If not have, return a empty dict.

    Returns:
        (dict[str,str]): extra cmds dict.
    """
    import imp

    extra_cmd_path: str = PIGIT_HOME + "/extra_cmds.py"
    extra_cmds = {}

    if os.path.isfile(extra_cmd_path):
        try:
            extra_cmd = imp.load_source("extra_cmd", extra_cmd_path)
        except Exception as e:
            Log.error(
                "Can't load file '{0}';{1};{2}".format(
                    extra_cmd_path, str(e), str(e.__traceback__)
                )
            )
        else:
            try:
                extra_cmds = extra_cmd.extra_cmds  # type: ignore
            except AttributeError:
                Log.error("Can't found dict name is 'extra_cmds'.")

    # print(extra_cmds)
    return extra_cmds


def shell_mode(git_processor: CmdProcessor):
    import pigit.tomato

    print(
        "Welcome come PIGIT shell.\n"
        "You can use short commands directly. Input '?' to get help.\n"
    )
    while True:
        command = input("(pigit)> ").strip()
        if not command:
            continue

        command_args = command.split()
        head_command = command_args[0]

        if head_command in ["quit", "exit"]:  # ctrl+c
            break
        elif head_command in git_processor.cmds.keys():
            git_processor.process_command(head_command, command_args[1:])
        elif head_command == "tomato":
            # Tomato clock.
            pigit.tomato.main(command_args)
        elif head_command == "?":
            other_ = command_args[1:]
            if not other_:
                git_processor.command_help()
            else:
                print(other_)
                for item in other_:
                    if item in git_processor.cmds.keys():
                        print(git_processor._generate_help_by_key(item))
                    elif item == "tomato":
                        pigit.tomato.help("tomato")
        else:
            os.system(command)
    return None


def interactive_interface(args):
    flag: int = 1

    if args:
        ready_flag = args[0]
        if ready_flag in "0123456789":
            flag = int(ready_flag)

    while flag and flag > 0:
        if flag == 1:
            flag = InteractiveStatus(
                use_color=CONFIG.gitprocessor_interactive_color,
                help_wait=CONFIG.gitprocessor_interactive_help_showtime,
            ).run()
        elif flag == 2:
            flag = InteractiveCommit(
                use_color=CONFIG.gitprocessor_interactive_color,
                help_wait=CONFIG.gitprocessor_interactive_help_showtime,
            ).run()
        else:
            flag = 1


class CustomHelpFormatter(argparse.HelpFormatter):
    """Formatter for generating usage messages and argument help strings.

    This class inherits `argparse.HelpFormatter` and rewrites some methods
    to complete customization.
    """

    def __init__(
        self, prog, indent_increment=2, max_help_position=24, width=90, colors=None
    ):
        width = CONFIG.help_max_line_width
        max_width, _ = get_terminal_size()

        width = width if width < max_width else max_width - 2
        super(CustomHelpFormatter, self).__init__(
            prog, indent_increment, max_help_position, width
        )
        if not colors or not isinstance(colors, list):
            colors = {
                "red": Color.fg("#FF6347"),  # Tomato
                "green": Color.fg("#98FB98"),  # PaleGreen
                "yellow": Color.fg("#EBCB8C"),  # Yellow
                "blue": Color.fg("#87CEFA"),  # SkyBlue
                # Color.fg("#FFC0CB"),  # Pink
            }
        self.colors = colors
        self.color_len = len(colors)
        self._old_color = None

    def _format_action(self, action):
        # determine the required width and the entry label
        help_position = min(self._action_max_length + 2, self._max_help_position)
        help_width = max(self._width - help_position, 11)
        action_width = help_position - self._current_indent - 2
        action_header = self._format_action_invocation(action)

        # no help; start on same line and add a final newline
        if not action.help:
            tup = self._current_indent, "", action_header
            action_header = "%*s%s\n" % tup

        # short action name; start on the same line and pad two spaces
        elif len(action_header) <= action_width:
            tup = self._current_indent, "", action_width, action_header
            action_header = "%*s%-*s  " % tup
            indent_first = 0

        # long action name; start on the next line
        else:
            tup = self._current_indent, "", action_header
            action_header = "%*s%s\n" % tup
            indent_first = help_position

        # collect the pieces of the action help
        # @Overwrite
        if CONFIG.help_use_color:
            _color = self.colors["green"]
        else:
            _color = ""
        parts = [_color + action_header + Fx.reset]

        # if there was help for the action, add lines of help text
        if action.help:
            help_text = self._expand_help(action)
            help_lines = self._split_lines(help_text, help_width)
            parts.append("%*s%s\n" % (indent_first, "", help_lines[0]))
            for line in help_lines[1:]:
                parts.append("%*s%s\n" % (help_position, "", line))

        # or add a newline if the description doesn't end with one
        elif not action_header.endswith("\n"):
            parts.append("\n")

        # if there are any sub-actions, add their help as well
        for subaction in self._iter_indented_subactions(action):
            parts.append(self._format_action(subaction))

        # return a single string
        return self._join_parts(parts)

    def _fill_text(self, text, width, indent):
        try:
            return "".join(indent + line for line in text.splitlines(keepends=True))
        except TypeError:
            return "".join(indent + line for line in text.splitlines())

    def _get_help_string(self, action):
        help = action.help
        if "%(default)" not in action.help:
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    # help += " (default: %(default)s)"
                    pass
        return help


class Parser(object):
    def __init__(self) -> None:
        super(Parser, self).__init__()
        self._parser = argparse.ArgumentParser(
            prog="pigit",
            description="If you want to use some original git commands, "
            "please use -- to indicate.",
            prefix_chars="-",
            formatter_class=CustomHelpFormatter,
        )
        self._add_arguments()

    def _add_arguments(self) -> None:
        p = self._parser

        p.add_argument(
            "-v",
            "--version",
            action="version",
            help="Show version and exit.",
            version="Version: %s" % __version__,
        )
        p.add_argument(
            "-C",
            "--complete",
            action="store_true",
            help="Add shell prompt script and exit.(Supported `bash`, `zsh`, `fish`)",
        )
        p.add_argument(
            "-s",
            "--show-commands",
            action="store_true",
            help="List all available short command and wealth and exit.",
        )
        p.add_argument(
            "-p",
            "--show-part-command",
            type=str,
            metavar="TYPE",
            dest="command_type",
            help="According to given type [%s] list available short command and "
            "wealth and exit." % ", ".join(CommandType.__members__.keys()),
        )
        p.add_argument(
            "-t",
            "--types",
            action="store_true",
            help="List all command types and exit.",
        )
        p.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="Current runtime in debug mode.",
        )
        p.add_argument(
            "--out-log",
            action="store_true",
            help="Print log to console.",
        )

        tool_group = p.add_argument_group(
            title="tools arguments",
            description="Auxiliary type commands.",
        )
        tool_group.add_argument(
            "-f",
            "--config",
            action="store_true",
            help="Display the config of current git repository and exit.",
        )
        tool_group.add_argument(
            "-i",
            "--information",
            action="store_true",
            help="Show some information about the current git repository.",
        )
        tool_group.add_argument(
            "-c",
            "--count",
            nargs="?",
            const=".",
            type=str,
            metavar="PATH",
            help="Count the number of codes and output them in tabular form."
            "A given path can be accepted, and the default is the current directory.",
        )
        tool_group.add_argument(
            "--create-ignore",
            type=str,
            metavar="TYPE",
            dest="ignore_type",
            help="Create a demo `.gitignore` file. Need one argument, support: [%s]"
            % ", ".join(GitignoreGenetor.Supported_Types.keys()),
        )
        p.add_argument(
            "--create-config",
            action="store_true",
            help="Create a preconfigured file of PIGIT."
            "(If a profile exists, the values available in it are used)",
        )
        tool_group.add_argument(
            "--shell",
            action="store_true",
            help="Go to the pigit shell mode.",
        )

        p.add_argument(
            "command", nargs="?", type=str, help="Short git command or other."
        )
        p.add_argument("args", nargs="*", type=str, help="Command parameter list.")

    def parse(
        self, custom_commands: Union[list, str, None] = None
    ) -> tuple[argparse.Namespace, list]:
        if custom_commands:
            if isinstance(custom_commands, list):
                args, unknown = self._parser.parse_known_args(custom_commands)
            elif isinstance(custom_commands, str):
                args, unknown = self._parser.parse_known_args(custom_commands.split())
            else:
                raise AttributeError("custom_commands need be list or str.")
        else:
            args, unknown = self._parser.parse_known_args()

        Log.debug("Parser result: {0}, {1}".format(args, unknown))
        return args, unknown

    def process(self, known_args, extra_unknown: Optional[list] = None) -> None:
        try:
            if known_args.config:
                output_git_local_config(CONFIG.git_config_format)

            if known_args.information:
                output_repository_info(
                    show_path=CONFIG.repository_show_path,
                    show_remote=CONFIG.repository_show_remote,
                    show_branches=CONFIG.repository_show_branchs,
                    show_lastest_log=CONFIG.repository_show_lastest_log,
                    show_summary=CONFIG.repository_show_summary,
                )

            if known_args.complete:
                # Generate competion vars dict.
                completion_vars = {
                    key: value.get("help", "") for key, value in Git_Cmds.items()
                }

                # Update var dict with shell command.
                completion_vars.update(process_argparse(self._parser))

                ShellCompletion(
                    __project__, completion_vars, PIGIT_HOME
                ).complete_and_use()
                return None

            if known_args.create_config:
                return CONFIG.create_config_template()

            if known_args.ignore_type:
                GitignoreGenetor(timeout=CONFIG.gitignore_generator_timeout,).launch(
                    known_args.ignore_type,
                    dir_path=REPOSITORY_PATH,
                )
                return None

            if known_args.count:
                path = known_args.count if known_args.count != "." else os.getcwd()
                CodeCounter(
                    count_path=path,
                    use_ignore=CONFIG.codecounter_use_gitignore,
                    result_saved_path=COUNTER_PATH,
                    result_format=CONFIG.codecounter_result_format,
                    use_icon=CONFIG.codecounter_show_icon,
                ).count_and_format_print(
                    show_invalid=CONFIG.codecounter_show_invalid,
                )
                return None

            extra_cmd = {
                "ui": {
                    "belong": CommandType.Index,
                    "command": interactive_interface,
                    "help": "interactive operation git tree status.",
                    "type": "func",
                    "has_arguments": True,
                },
                "shell": {
                    "command": "",
                    "help": "Into PIGIT shell mode.",
                },  # only for tips.
            }
            extra_cmd.update(get_extra_cmds())

            git_processor = CmdProcessor(
                extra_cmds=extra_cmd,
                use_recommend=CONFIG.gitprocessor_use_recommend,
                show_original=CONFIG.gitprocessor_show_original,
                use_color=CONFIG.gitprocessor_interactive_color,
                help_wait=CONFIG.gitprocessor_interactive_help_showtime,
            )

            if known_args.shell:
                shell_mode(git_processor)

            if known_args.show_commands:
                return git_processor.command_help()

            if known_args.command_type:
                return git_processor.command_help_by_type(known_args.command_type)

            if known_args.types:
                return git_processor.type_help()

            if known_args.command:
                if known_args.command == "shell":
                    return shell_mode(git_processor)
                else:
                    command = known_args.command
                    known_args.args.extend(extra_unknown)  # type: list
                    git_processor.process_command(command, known_args.args)
                    return None

            # Don't have invalid command list.
            if not list(filter(lambda x: x, vars(known_args).values())):
                introduce()
        except (KeyboardInterrupt, EOFError):
            raise SystemExit(0)


@time_it
def main(custom_commands: Optional[list] = None):
    parser = Parser()

    # Parse custom comand or parse input command.
    if custom_commands is not None:
        stdargs, extra_unknown = parser.parse(custom_commands)
    stdargs, extra_unknown = parser.parse()

    # Setup log handle.
    log_file = LOG_PATH if stdargs.out_log or CONFIG.stream_output_log else None
    LogHandle.setup_logging(debug=stdargs.debug or CONFIG.debug_mode, log_file=log_file)

    # Process result.
    parser.process(stdargs, extra_unknown)


if __name__ == "__main__":
    main()



import pytest
from .utils import analyze_it

from pigit.processor import CmdProcessor
from pigit.common import Fx


def test_init():
    with pytest.raises(TypeError):
        CmdProcessor(extra_cmds="xxx")


@pytest.fixture(scope="module")
def setup():
    extra = {"aa": {"help": "print system user name."}}
    return CmdProcessor(extra_cmds=extra)


def test_process_error(setup):
    handle = setup

    with pytest.raises(ValueError):
        handle._generate_help_by_key("aa")


@pytest.mark.parametrize(
    "command",
    [
        "git status",
        "git log --online --graph",
        "git log --dep 10 --online --graph",
        "git log --dep 10 --online --graph --color true",
        "git add xxx/xxx",
        "git checkout -b test",
    ],
)
def test_color_command(setup, command: str):
    handle = setup

    color_str = handle.color_command(command)
    # print(repr(color_str))
    print(color_str)
    assert Fx.pure(color_str).strip() == command.strip()



# -*- coding:utf-8 -*-

from ..common import Fx, Color, shorten
from .base import _Interaction
from .model import Commit


class InteractiveShowCommitInfo(_Interaction):
    def keyevent_help(self) -> str:
        return ""

    def process_keyevent(self, input_key: str, cursor_row: int, data) -> bool:
        pass

    def get_raw_data(self) -> list[str]:
        return self._data_handle.load_commit_info(
            self._ex_commit.sha,
        ).split("\n")

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print("{}{}{}".format(Color.bg("#6495ED"), line, Fx.reset))
        else:
            print(line)


class InteractiveCommit(_Interaction):
    def keyevent_help(self) -> str:
        return "↲ : check commit diff.\n"

    def process_keyevent(
        self, input_key: str, cursor_row: int, data: list[Commit]
    ) -> bool:
        if input_key == "enter":
            InteractiveShowCommitInfo(
                self.use_color,
                self.cursor,
                self.help_wait,
                True,
                self._debug,
                commit=data[cursor_row - 1],
            ).run()

    def get_raw_data(self) -> list[Commit]:
        current = self._data_handle.current_head()
        return self._data_handle.load_commits(current)

    def process_raw_data(
        self, raw_data: list[Commit], width: int
    ) -> list[tuple[str, int]]:
        color_data = []
        pushed_c = Color.fg("#F0E68C")
        unpushed_c = Color.fg("#F08080")
        for commit in raw_data:
            c = pushed_c if commit.is_pushed() else unpushed_c
            sha = commit.sha[:7]
            msg = shorten(commit.msg, width - 2)
            color_data.append(f"{c}{sha} {msg}{Fx.rs}")

        return super().process_raw_data(color_data, width)

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print("{} {}".format(self.cursor, line))
        else:
            print("  " + line)



import pytest
from .utils import analyze_it

from pigit import GitignoreGenetor


@analyze_it
def test_fetch_gitignore():
    base_url = "https://github.com/github/gitignore/blob/master/%s.gitignore"
    name = "Python"

    handle = GitignoreGenetor()
    content = handle.get_html_from_url(base_url % name)
    assert not content or "<!DOCTYPE html>" in content

    if content:
        ignore_content = handle.parse_gitignore_page(content)
        print(ignore_content)



# -*- coding:utf-8 -*-

import re
import time
from math import ceil
from abc import ABC, abstractmethod
from shutil import get_terminal_size
from typing import Optional, Any

from ..common import Fx, TermColor, exec_cmd, shorten, get_width, color_print
from ..common.singleton import Singleton
from ..keyevent import get_keyevent_obj
from ..gitinfo import REPOSITORY_PATH
from .model import File, Commit


class InteractionError(Exception):
    pass


class DataHandle(object, metaclass=Singleton):
    def __init__(self, use_color: bool):
        self.use_color = use_color

    def current_head(self):
        """Get current repo head.

        return a branch name or a commit sha string.
        """
        _, res = exec_cmd("git symbolic-ref -q --short HEAD")
        return res.rstrip()

    def load_status(self, max_width: int, ident: int = 2) -> list[File]:
        """Get the file tree status of GIT for processing and encapsulation.

        Args:
            max_width (int): The max length of display string.
            ident (int, option): Number of reserved blank characters in the header.

        Raises:
            Exception: Can't get tree status.

        Returns:
            (list[File]): Processed file status list.
        """

        file_items = []
        err, files = exec_cmd("git status -s -u --porcelain")
        if err:
            raise Exception("Can't get git status.")
        for file in files.rstrip().split("\n"):
            if not file.strip():
                # skip blank line.
                continue
            change = file[:2]
            staged_change = file[:1]
            unstaged_change = file[1:2]
            name = file[3:]
            untracked = change in ["??", "A ", "AM"]
            has_no_staged_change = staged_change in [" ", "U", "?"]
            has_merged_conflicts = change in ["DD", "AA", "UU", "AU", "UA", "UD", "DU"]
            has_inline_merged_conflicts = change in ["UU", "AA"]

            display_name = shorten(name, max_width - 3 - ident)
            # color full command.
            if unstaged_change != " ":
                if not has_no_staged_change:
                    display_str = "{}{}{}{} {}{}".format(
                        TermColor.Green,
                        staged_change,
                        TermColor.Red,
                        unstaged_change,
                        display_name,
                        Fx.reset,
                    )
                else:
                    display_str = "{}{} {}{}".format(
                        TermColor.Red, change, display_name, Fx.reset
                    )
            else:
                display_str = "{}{} {}{}".format(
                    TermColor.Green, change, display_name, Fx.reset
                )

            file_ = File(
                name=name,
                display_str=display_str if self.use_color else file,
                short_status=change,
                has_staged_change=not has_no_staged_change,
                has_unstaged_change=unstaged_change != " ",
                tracked=not untracked,
                deleted=unstaged_change == "D" or staged_change == "D",
                added=unstaged_change == "A" or untracked,
                has_merged_conflicts=has_merged_conflicts,
                has_inline_merged_conflicts=has_inline_merged_conflicts,
            )
            file_items.append(file_)

        return file_items

    def load_file_diff(
        self, file: str, tracked: bool = True, cached: bool = False, plain: bool = False
    ) -> str:
        """Gets the modification of the file.

        Args:
            file (str): file path relative to git.
            tracked (bool, optional): Defaults to True.
            cached (bool, optional): Defaults to False.
            plain (bool, optional): Whether need color. Defaults to False.

        Returns:
            (str): change string.
        """

        command = "git diff --submodule --no-ext-diff {plain} {cached} {tracked} {file}"

        if plain:
            _plain = "--color=never"
        else:
            _plain = "--color=always"

        if cached:
            _cached = "--cached"
        else:
            _cached = ""

        if not tracked:
            _tracked = "--no-index -- /dev/null"
        else:
            _tracked = "--"

        if "->" in file:  # rename status.
            file = file.split("->")[-1].strip()

        err, res = exec_cmd(
            command.format(plain=_plain, cached=_cached, tracked=_tracked, file=file)
        )
        if err:
            return "Can't get diff."
        return res.rstrip()

    def load_commits(
        self, branch_name: str, limit: bool = True, filter_path: str = ""
    ) -> list[Commit]:
        """Get the all commit of a given branch.

        Args:
            branch_name (str): want branch name.
            limit (bool): Whether to get only the latest 300.
            filter_path (str): filter dir path, default is empty.
        """

        passed_first_pushed_commit = False
        command = "git merge-base %s %s@{u}" % (branch_name, branch_name)
        _, resp = exec_cmd(command)
        first_pushed_commit = resp.strip()

        if not first_pushed_commit:
            passed_first_pushed_commit = True

        commits: list[Commit] = []

        # Generate git command.
        limit_flag = "-300" if limit else ""
        filter_flag = f"--follow -- {filter_path}" if filter_path else ""
        command = f'git log {branch_name} --oneline --pretty=format:"%H|%at|%aN|%d|%p|%s" {limit_flag} --abbrev=20 --date=unix {filter_flag}'
        err, resp = exec_cmd(command)

        # Process data.
        lines = resp.split("\n")
        if not err:
            for line in lines:
                split_ = line.split("|")

                sha = split_[0]
                unix_timestamp = int(split_[1])
                author = split_[2]
                extra_info = (split_[3]).strip()
                # parent_hashes = split_[4]
                message = "|".join(split_[5:])

                tag = []
                if extra_info:
                    _re = re.compile(r"tag: ([^,\\]+)")
                    match = _re.search(extra_info)
                    if match:
                        tag.append(match[1])

                if sha == first_pushed_commit:
                    passed_first_pushed_commit = True
                status = {True: "unpushed", False: "pushed"}[
                    not passed_first_pushed_commit
                ]

                commit_ = Commit(
                    sha=sha,
                    msg=message,
                    author=author,
                    unix_timestamp=unix_timestamp,
                    status=status,
                    extra_info=extra_info,
                    tag=tag,
                )
                commits.append(commit_)

        return commits

    def load_commit_info(
        self, commit_sha: str, file_name: str = "", plain: bool = False
    ) -> str:
        """Gets the change of a file or all in a given commit.

        Args:
            commit_sha: commit id.
            file_name: file name(include full path).
            plain: whether has color.
        """

        color_str = "never" if plain else "always"

        command = "git show --color=%s %s %s" % (color_str, commit_sha, file_name)
        _, resp = exec_cmd(command)
        return resp.rstrip()


# whether already in interactive mode.
_Interaction_Starting = False


class _Interaction(ABC):
    def __init__(
        self,
        use_color: bool = True,
        cursor: Optional[str] = None,
        help_wait: float = 1.5,
        is_sub: bool = False,  # whether is sub page.
        debug: bool = False,
        **kwargs,
    ) -> None:
        self.use_color = use_color
        self._is_sub = is_sub

        if not cursor:
            self.cursor = "→"
        elif len(cursor) == 1:
            if get_width(ord(cursor)) == 1:
                self.cursor = cursor
            else:
                self.cursor = "→"
                print("When displayed, it will occupy more than one character.")
        else:
            self.cursor = "→"
            print("The cursor symbol entered is not supported.")

        self.help_wait = help_wait
        self._min_height = 8
        self._min_width = 60
        self._debug = debug

        # Whether can into interactive.
        try:
            _keyevent_class = get_keyevent_obj()
        except NameError:
            raise InteractionError(
                "This behavior is not supported in the current system."
            )
        self._keyevent = _keyevent_class()

        self._data_handle = DataHandle(use_color)

        for key, value in kwargs.items():
            setattr(self, "_ex_{}".format(key), value)

    @abstractmethod
    def process_keyevent(self, input_key: str, cursor_row: int, data: Any) -> bool:
        """Handles keyboard events other than movement.

        Args:
            input_key (str): keyboard string.
            cursor_row (int): current line.
            data (Any): raw data.

        Returns:
            bool: whether need refresh data.
        """

    @abstractmethod
    def keyevent_help(self) -> str:
        """Get extra keyevent help message.

        Returns:
            str: help message string.
        """

    @abstractmethod
    def get_raw_data(self) -> list[Any]:
        """How to get the raw data."""

    def process_raw_data(
        self, raw_data: list[str], width: int
    ) -> list[tuple[str, int]]:
        new_list = []
        for line in raw_data:
            text = Fx.uncolor(line)
            count = 0
            for ch in text:
                count += get_width(ord(ch))
            # [float] is to solve the division of python2 without
            # retaining decimal places.
            new_list.append((line, ceil(count / width) - 1))
        return new_list

    @abstractmethod
    def print_line(self, line: str, is_cursor_row: bool) -> None:
        """How to output one line.

        May has some different when current line is cursor line.
        Support to process cursor line specially.
        """

    def run(self, *args):
        """Run method."""
        global _Interaction_Starting

        # Make sure run at a git repo dir.
        if not REPOSITORY_PATH:
            color_print("Current path is not a git repository.", TermColor.Red)
            return

        self.width, self.height = width, height = get_terminal_size()

        # Make sure have enough space.
        if height < self._min_height or width < self._min_width:
            raise InteractionError(
                "The minimum size of terminal should be {0} x {1}.".format(
                    self._min_width, self._min_height
                )
            )

        if self._debug:  # debug show.
            print(Fx.clear_)
            print(width, height)
            time.sleep(1.5)

        # Initialize.
        cursor_row: int = 1
        display_range: list = [1, height - 1]  # allow display row range.

        stopping: bool = False  # Stop flag.
        refresh: bool = False  # refresh data flag.

        raw_data: list[Any] = self.get_raw_data()
        self.raw_data = raw_data
        if not raw_data:
            print("The work tree is clean and there is nothing to operate.")
            return
        show_data = self.process_raw_data(raw_data, width)

        extra = 0  # Extra occupied row.

        # Into new term page.
        try:
            if not _Interaction_Starting:
                _Interaction_Starting = True
                print(Fx.alt_screen + Fx.hide_cursor)
                #  try hook window resize event.
                self._keyevent.signal_init()

            # Start interactive.
            while not stopping:
                if refresh:
                    refresh = False
                    raw_data: list[Any] = self.get_raw_data()
                    self.raw_data = raw_data
                    show_data = self.process_raw_data(raw_data, width)

                print(Fx.clear_)

                # check whether have status.
                if not raw_data:
                    print("The work tree is clean and there is nothing to operate.")
                    time.sleep(1)
                    return

                while cursor_row < display_range[0]:
                    display_range = [i - 1 for i in display_range]
                while cursor_row + extra > display_range[1]:
                    display_range = [i + 1 for i in display_range]

                extra = 0  # Return to zero and accumulate again.

                # Print needed display part.
                for index, item in enumerate(show_data, start=1):
                    line, each_extra = item
                    if display_range[0] <= index <= display_range[1] - extra:
                        self.print_line(line, index == cursor_row)
                        extra += each_extra

                input_key = self._keyevent.sync_get_input()
                if input_key in ["q", "escape"]:
                    # exit.
                    stopping = True
                elif input_key in ["j", "down"]:
                    # select pre file.
                    cursor_row += 1
                    cursor_row = min(cursor_row, len(show_data))
                elif input_key in ["k", "up"]:
                    # select next file.
                    cursor_row -= 1
                    cursor_row = max(cursor_row, 1)
                elif input_key in ["J"]:
                    # scroll down 5 lines.
                    cursor_row += 5
                    cursor_row = min(cursor_row, len(show_data))
                elif input_key in ["K"]:
                    # scroll up 5 line
                    cursor_row -= 5
                    cursor_row = max(cursor_row, 1)

                elif input_key in "0123456789":
                    return int(input_key)
                elif input_key == "windows resize":
                    # get new term height.
                    new_width, new_height = get_terminal_size()
                    if new_height < self._min_height or new_width < self._min_width:
                        raise InteractionError(
                            "The minimum size of terminal should be {0} x {1}.".format(
                                self._min_width, self._min_height
                            )
                        )
                    # get diff, reassign.
                    line_diff = new_height - height
                    width, height = new_width, new_height
                    # get new display range.
                    display_range[1] += line_diff
                    show_data = self.process_raw_data(raw_data, width)
                elif input_key in ["?", "h"]:
                    print(Fx.clear_)
                    print(
                        (
                            "k / ↑: select previous line.\n"
                            "j / ↓: select next line.\n"
                            "J: Scroll down 5 lines.\n"
                            "K: Scroll down 5 lines.\n"
                            "? / h : show help, wait {}s and exit.\n"
                            + self.keyevent_help()
                        ).format(self.help_wait)
                    )
                    if self.help_wait == 0:
                        self._keyevent.sync_get_input()
                    else:
                        time.sleep(self.help_wait)
                else:
                    refresh = self.process_keyevent(input_key, cursor_row, raw_data)

        except KeyboardInterrupt:
            pass
        finally:
            # Whatever, unregister signal event and restore terminal at last.
            if _Interaction_Starting and not self._is_sub:
                _Interaction_Starting = False
                self._keyevent.signal_restore()
                print(Fx.normal_screen + Fx.show_cursor, end="")



# -*- coding:utf-8 -*-

import os

from ..common import Fx, Color, exec_cmd, run_cmd, confirm
from .base import _Interaction, InteractionError
from .model import File


class InteractiveShowFile(_Interaction):
    def keyevent_help(self) -> str:
        return ""

    def process_keyevent(self):
        pass

    def get_raw_data(self) -> list[str]:
        return self._data_handle.load_file_diff(
            self._ex_file.name,
            self._ex_file.tracked,
            self._ex_file.has_staged_change,
            not self.use_color,
        ).split("\n")

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print("{}{}{}".format(Color.bg("#6495ED"), line, Fx.reset))
        else:
            print(line)


class InteractiveStatus(_Interaction):
    """Interactive operation git tree status."""

    def keyevent_help(self) -> str:
        return (
            "a / space: toggle storage or unstorage file.\n"
            "d: discard the file changed.\n"
            "e: open file with default editor.\n"
            "↲ : check file diff.\n"
        )

    def process_keyevent(
        self, input_key: str, cursor_row: int, data: list[File]
    ) -> bool:
        refresh = False
        if input_key in ["a", "space"]:
            self.process_f(data[cursor_row - 1], "switch")
            refresh = True
        elif input_key == "d":
            self.process_f(data[cursor_row - 1], "discard")
            refresh = True
        elif input_key == "e":
            editor = os.environ.get("EDITOR", None)
            if editor:
                run_cmd('{} "{}"'.format(editor, data[cursor_row - 1].name))
                refresh = True
            else:
                # No default editor to open file.
                pass
        elif input_key == "enter":
            InteractiveShowFile(
                self.use_color,
                self.cursor,
                self.help_wait,
                True,
                self._debug,
                file=data[cursor_row - 1],
            ).run()

        return refresh

    def process_f(self, file: File, flag: str) -> None:
        """Process file to change the status.

        Args:
            file (File): One processed file.
        """

        if flag == "switch":
            if file.has_merged_conflicts or file.has_inline_merged_conflicts:
                pass
            elif file.has_unstaged_change:
                exec_cmd("git add -- {}".format(file.name))
            elif file.has_staged_change:
                if file.tracked:
                    exec_cmd("git reset HEAD -- {}".format(file.name))
                else:
                    exec_cmd("git rm --cached --force -- {}".format(file.name))

        elif flag == "discard":
            print(Fx.clear_)
            if confirm("discard all changed? [y/n]:"):
                if file.tracked:
                    exec_cmd("git checkout -- {}".format(file.name))
                else:
                    os.remove(os.path.join(file.name))

    def get_raw_data(self) -> list[File]:
        return self._data_handle.load_status(self.width)

    def process_raw_data(
        self, raw_data: list[File], width: int
    ) -> list[tuple[str, int]]:
        l = [file.display_str for file in raw_data]
        return super().process_raw_data(l, width)

    def print_line(self, line: str, is_cursor_row: bool) -> None:
        if is_cursor_row:
            print("{} {}".format(self.cursor, line))
        else:
            print("  " + line)



# -*- coding:utf-8 -*-

import re
import textwrap
import logging
from typing import Optional

from .common import Fx, TermColor, exec_cmd, color_print
from .common.git_utils import parse_git_config, git_version, current_repository
from .common.str_table import dTable, TableTooWideError

Log = logging.getLogger(__name__)


# Not detected, the result is empty str.
Git_Version: str = git_version()

# Not a repository, the all path is empty str.
REPOSITORY_PATH, GIT_CONF_PATH = current_repository()


def _config_normal_output(conf: dict[str, dict]) -> None:
    for t, d in conf.items():
        print(TermColor.Red + f"[{t}]")
        for k, v in d.items():
            print(" " * 4 + TermColor.SkyBlue + k, end="")
            print("=" + Fx.italic + TermColor.MediumVioletRed + v + Fx.reset)


def _config_table_output(conf: dict[str, dict]) -> None:
    for sub in conf.values():
        for k, v in sub.items():
            sub[k] = f"{TermColor.Green}{v:<40}{Fx.rs}"

    tb = dTable(conf, title="Git Local Config")
    tb.print()


output_way = {
    "normal": _config_normal_output,
    "table": _config_table_output,
}


def output_git_local_config(style: str = "table") -> None:
    """Print the local config of current git repository."""

    if not REPOSITORY_PATH:
        color_print("This directory is not a git repository yet.", TermColor.Red)
        return None

    try:
        with open(GIT_CONF_PATH + "/config", "r") as cf:
            context = cf.read()
    except Exception as e:
        color_print(
            "Error reading configuration file. {0}".format(str(e)), TermColor.Red
        )
    else:
        config_dict = parse_git_config(context)

        try:
            output_way[style](config_dict)
        except (KeyError, TableTooWideError) as e:
            # There are two different causes of errors that can be triggered here.
            # First, a non-existent format string is passed in (theoretically impossible),
            # but terminal does not have enough width to display the table.
            output_way["normal"](config_dict)

            # log error info.
            Log.error(str(e) + str(e.__traceback__))


def output_repository_info(
    show_path: bool = True,
    show_remote: bool = True,
    show_branches: bool = True,
    show_lastest_log: bool = True,
    show_summary: bool = True,
) -> None:
    """Print some information of the repository.

    repository: `Repository_Path`
    remote: read from '.git/conf'
    >>> all_branch = run_cmd_with_resp('git branch --all --color')
    >>> lastest_log = run_cmd_with_resp('git log -1')
    """

    print("waiting ...", end="")

    error_str = TermColor.Red + "Error getting" + Fx.reset

    # Print content.
    print("\r[%s]        \n" % (Fx.b + "Repository Information" + Fx.reset,))
    if show_path:
        print(
            "Repository: \n\t%s\n" % (TermColor.SkyBlue + REPOSITORY_PATH + Fx.reset,)
        )

    # Get remote url.
    if show_remote:
        try:
            with open(REPOSITORY_PATH + "/.git/config", "r") as cf:
                config = cf.read()
        except Exception:
            remote = error_str
        else:
            res = re.findall(r"url\s=\s(.*)", config)
            remote = "\n".join(
                [
                    "\t%s%s%s%s" % (Fx.italic, TermColor.SkyBlue, x, Fx.reset)
                    for x in res
                ]
            )
        print("Remote: \n%s\n" % remote)

    # Get all branches.
    if show_branches:
        err, res = exec_cmd("git branch --all --color")
        if err:
            branches = "\t" + error_str
        else:
            branches = textwrap.indent(res, "\t")
        print("Branches: \n%s\n" % branches)

    # Get the lastest log.
    if show_lastest_log:
        err, res = exec_cmd("git log --stat --oneline --decorate -1 --color")
        if err:
            git_log = "\t" + error_str
        else:
            # git_log = "\n".join(["\t" + x for x in res.strip().split("\n")])
            git_log = textwrap.indent(res, "\t")
        print("Lastest log:\n%s\n" % git_log)

    # Get git summary.
    if show_summary:
        err, res = exec_cmd("git shortlog --summary --numbered")
        if err:
            summary = "\t" + error_str
        else:
            summary = textwrap.indent(res, "\t")
        print("Summary:\n%s\n" % summary)
